try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import models, serializers


def test_save(password_reset_token_factory):
    """
    Saving the serializer with a valid token and password should reset
    the associated user's password.
    """
    token = password_reset_token_factory()
    user = token.email.user

    data = {
        'key': token.key,
        'password': 'new_passw0rd',
    }

    serializer = serializers.PasswordResetSerializer(data=data)
    assert serializer.is_valid()

    serializer.save()
    user.refresh_from_db()

    assert user.check_password(data['password'])
    assert models.PasswordResetToken.objects.count() == 0


def test_validate_key_expired(password_reset_token_factory):
    """
    If the key provided corresponds to a token that has expired, the
    serializer should not validate.
    """
    token = password_reset_token_factory()
    data = {
        'key': token.key,
        'password': 'new_passw0rd',
    }

    serializer = serializers.PasswordResetSerializer(data=data)

    with mock.patch(
            'rest_email_auth.serializers.models.PasswordResetToken.valid_tokens.filter',   # noqa
            return_value=models.PasswordResetToken.objects.none()) as mock_filter:         # noqa
        assert not serializer.is_valid()

    assert set(serializer.errors.keys()) == {'key'}
    assert mock_filter.call_count == 1


def test_validate_key_invalid(db):
    """
    If there is no reset token with the provided key, the serializer
    should not validate.
    """
    data = {
        'key': '00000000-0000-0000-0000-000000000000',
        'password': 'new_passw0rd',
    }

    serializer = serializers.PasswordResetSerializer(data=data)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'key'}


@mock.patch(
    'rest_email_auth.serializers.password_validation.validate_password',
    autospec=True)
def test_validate_password(mock_validate_password):
    """
    The serializer should pass the provided password through Django's
    password validation system.
    """
    data = {
        'key': 'foo',
        'password': 'foobar',
    }
    serializer = serializers.PasswordResetSerializer(data=data)

    serializer.is_valid()

    assert mock_validate_password.call_count == 1
    assert mock_validate_password.call_args[0] == (data['password'],)



from rest_framework import serializers, status

from rest_email_auth import generics


class ExampleSerializer(serializers.Serializer):
    """
    Serializer for testing views.
    """
    foo = serializers.CharField()

    def save(self):
        """
        Mock save method.
        """
        pass


class ExampleView(generics.SerializerSaveView):
    """
    View that extends from the generic view we're testing.
    """
    serializer_class = ExampleSerializer


def test_post_invalid(api_rf):
    """
    Sending an invalid POST request to the view should return the
    serializer's errors.
    """
    request = api_rf.post('/', {})
    response = ExampleView.as_view()(request)

    assert response.status_code == status.HTTP_400_BAD_REQUEST

    serializer = ExampleSerializer(data={})
    assert not serializer.is_valid()

    assert response.data == serializer.errors


def test_post_valid(api_rf):
    """
    Sending a valid POST request to the view should call the
    serializer's ``save`` method and return the serializer's data.
    """
    data = {
        'foo': 'bar',
    }

    request = api_rf.post('/', data)
    response = ExampleView.as_view()(request)

    assert response.status_code == status.HTTP_200_OK

    serializer = ExampleSerializer(data=data)
    assert serializer.is_valid()

    assert response.data == serializer.data



"""Factories for generating test data.

These factories generate models with dummy data for testing purposes.
"""

from django.contrib.auth import get_user_model

import factory

from rest_email_auth import models


class EmailConfirmationFactory(factory.django.DjangoModelFactory):
    """
    Factory for generating email confirmations.
    """
    email = factory.SubFactory('rest_email_auth.factories.EmailFactory')

    class Meta(object):
        model = models.EmailConfirmation


class EmailFactory(factory.django.DjangoModelFactory):
    """
    Factory for generating email addresses.
    """
    email = factory.sequence(lambda n: 'test{n}@example.com'.format(n=n))
    user = factory.SubFactory('rest_email_auth.factories.UserFactory')

    class Meta(object):
        model = models.EmailAddress


class PasswordResetTokenFactory(factory.django.DjangoModelFactory):
    """
    Factory for generating password reset tokens.
    """
    email = factory.SubFactory('rest_email_auth.factories.EmailFactory')

    class Meta(object):
        model = 'rest_email_auth.PasswordResetToken'


class UserFactory(factory.django.DjangoModelFactory):
    """
    Factory for generating Django user instances.
    """
    password = 'password'
    username = factory.sequence(lambda n: 'user{n}'.format(n=n))

    class Meta(object):
        model = get_user_model()

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        """
        Create a new user instance.

        Args:
            model_class:
                The type of model to create an instance of.
            args:
                Positional arguments to create the instance with.
            kwargs:
                Keyword arguments to create the instance with.

        Returns:
            A new user instance of the type specified by
            ``model_class``.
        """
        manager = cls._get_manager(model_class)

        return manager.create_user(*args, **kwargs)



from rest_framework import status
from rest_framework.reverse import reverse

from rest_email_auth import serializers


url = reverse('rest-email-auth:email-list')


def test_create_email(api_client, user_factory):
    """
    Sending a POST request with valid data to the view should create a
    new email address associated with the account.
    """
    user = user_factory()
    api_client.force_authenticate(user=user)

    data = {
        'email': 'test-add@example.com',
    }

    response = api_client.post(url, data)

    assert response.status_code == status.HTTP_201_CREATED

    serializer = serializers.EmailSerializer(user.email_addresses.get())

    assert response.data == serializer.data


def test_list_emails(api_client, email_factory, user_factory):
    """
    Sending a GET request to the view should return the requesting
    user's emails.
    """
    user = user_factory()
    email_factory(user=user)
    email_factory(user=user)

    # Create email address for other user
    email_factory()

    serializer = serializers.EmailSerializer(
        user.email_addresses.all(),
        many=True)

    api_client.force_authenticate(user=user)

    response = api_client.get(url)

    assert response.status_code == status.HTTP_200_OK
    assert response.data == serializer.data


def test_list_emails_anonymous(api_client):
    """
    Sending a GET request to the view as an anonymous user should result
    in a permissions error.
    """
    response = api_client.get(url)

    assert response.status_code == status.HTTP_403_FORBIDDEN



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# django-rest-email-auth documentation build configuration file, created
# by sphinx-quickstart on Fri Nov  3 12:44:08 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.

import os


# -- General configuration ---------------------------------------------


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinxcontrib.httpdomain',
    'sphinx_issues',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'django-rest-email-auth'
copyright = '2017, Chathan Driehuys'
author = 'Chathan Driehuys'

# The version info for the project you're documenting, acts as
# replacement for |version| and |release|, also used in various other
# places throughout the built documents.
#
# The short X.Y version.
version = '0.4.2'
# The full version, including alpha/beta/rc tags.
release = '0.4.2'

# The language for content autogenerated by Sphinx. Refer to
# documentation for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce
# nothing.
todo_include_todos = False


# Configuration for 'sphinx-issues'
issues_github_path = 'cdriehuys/django-rest-email-auth'


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.
if os.environ.get('READTHEDOCS') == 'True':
    html_theme = 'default'
else:
    html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = ['_static']


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'django-rest-email-authdoc'


# -- Options for LaTeX output ------------------------------------------

latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        'django-rest-email-auth.tex',
        'django-rest-email-auth Documentation',
        'Chathan Driehuys',
        'manual',
    ),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        'django-rest-email-auth',
        'django-rest-email-auth Documentation',
        [author],
        1,
    ),
]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'django-rest-email-auth',
        'django-rest-email-auth Documentation',
        author,
        'django-rest-email-auth',
        'One line description of project.',
        'Miscellaneous',
    ),
]



import datetime

from django.core import management
from django.utils.six import StringIO

import pytest

from rest_email_auth import models


def test_clean_expired(email_confirmation_factory, settings):
    """
    Calling the command should remove any email confirmations that are
    a specified time older than their expiration date.
    """
    settings.REST_EMAIL_AUTH = {
        'CONFIRMATION_EXPIRATION': datetime.timedelta(seconds=0),
        'CONFIRMATION_SAVE_PERIOD': datetime.timedelta(seconds=0),
    }

    email_confirmation_factory()

    out = StringIO()
    management.call_command('cleanemailconfirmations', stdout=out)

    assert models.EmailConfirmation.objects.count() == 0
    assert 'Removed 1' in out.getvalue()


def test_clean_just_expired(email_confirmation_factory, settings):
    """
    Email confirmations that have just expired should not be removed.
    """
    settings.REST_EMAIL_AUTH = {
        'CONFIRMATION_EXPIRATION': datetime.timedelta(seconds=0),
    }

    email_confirmation_factory()

    management.call_command('cleanemailconfirmations')

    assert models.EmailConfirmation.objects.count() == 1


@pytest.mark.django_db
def test_clean_none():
    """
    If there are no expired email confirmations, a message should be
    displayed indicating that nothing happened.
    """
    out = StringIO()
    management.call_command('cleanemailconfirmations', stdout=out)

    assert 'No email confirmations to remove' in out.getvalue()



"""
URL configuration for the ``rest_email_auth`` app.
"""

from django.conf.urls import url

from rest_email_auth import views


app_name = 'rest-email-auth'


urlpatterns = [
    url(
        r'^emails/$',
        views.EmailListView.as_view(),
        name='email-list'),

    url(r'^emails/(?P<pk>[0-9]+)/$',
        views.EmailDetailView.as_view(),
        name='email-detail'),

    url(
        r'^register/$',
        views.RegistrationView.as_view(),
        name='register'),

    url(
        r'^request-password-reset/$',
        views.PasswordResetRequestView.as_view(),
        name='password-reset-request'),

    url(
        r'^resend-verification/$',
        views.ResendVerificationView.as_view(),
        name='resend-verification'),

    url(
        r'^reset-password/$',
        views.PasswordResetView.as_view(),
        name='password-reset'),

    url(
        r'^verify-email/$',
        views.EmailVerificationView.as_view(),
        name='verify-email'),
]



from rest_framework import status

from rest_email_auth import serializers, views


email_verification_view = views.EmailVerificationView.as_view()


def test_verify_email(
        api_rf,
        email_confirmation_factory,
        email_factory,
        user_factory):
    """
    Sending a POST request with valid data to the endpoint should mark
    the associated email address as verified.
    """
    user = user_factory(password='password')
    email = email_factory(user=user)
    confirmation = email_confirmation_factory(email=email)

    data = {
        'key': confirmation.key,
        'password': 'password',
    }

    serializer = serializers.EmailVerificationSerializer(data=data)
    assert serializer.is_valid()

    request = api_rf.post('/', data)
    response = email_verification_view(request)

    assert response.status_code == status.HTTP_200_OK
    assert response.data == serializer.data



from django.core import management
from django.utils import timezone

from rest_email_auth import app_settings, models


class Command(management.BaseCommand):
    """
    Command to clean up old email confirmations.
    """
    help = "Clean up expired email confirmations."

    def handle(self, *args, **kwargs):
        """
        Handle execution of the command.
        """
        cutoff = timezone.now()
        cutoff -= app_settings.CONFIRMATION_EXPIRATION
        cutoff -= app_settings.CONFIRMATION_SAVE_PERIOD

        queryset = models.EmailConfirmation.objects.filter(
            created_at__lte=cutoff)

        count = queryset.count()

        queryset.delete()

        if count:
            self.stdout.write(
                self.style.SUCCESS(
                    'Removed {count} old email confirmation(s)'.format(
                        count=count)))
        else:
            self.stdout.write('No email confirmations to remove.')



"""
Settings specific to the application.

The setting implementation is modeled off of "Django Allauth's" from:
https://github.com/pennersr/django-allauth/blob/master/allauth/account/app_settings.py
"""

import sys


class AppSettings(object):

    def __init__(self):
        """
        Perform some basic settings checks.
        """
        # The user must provide a URL template for the verification
        # and password reset endpoints.
        assert self.EMAIL_VERIFICATION_URL
        assert self.PASSWORD_RESET_URL

    def _setting(self, name, default):
        """
        Retrieve a setting from the current Django settings.

        Settings are retrieved from the ``REST_EMAIL_AUTH`` dict in the
        settings file.

        Args:
            name (str):
                The name of the setting to retrieve.
            default:
                The setting's default value.

        Returns:
            The value provided in the settings dictionary if it exists.
            The default value is returned otherwise.
        """
        from django.conf import settings

        settings_dict = getattr(settings, 'REST_EMAIL_AUTH', {})

        return settings_dict.get(name, default)

    @property
    def CONFIRMATION_EXPIRATION(self):
        """
        The duration that an email confirmation is valid for.

        Defaults to 1 day.
        """
        import datetime

        return self._setting(
            'CONFIRMATION_EXPIRATION',
            datetime.timedelta(days=1))

    @property
    def CONFIRMATION_SAVE_PERIOD(self):
        """
        The duration that expired confirmations are saved for.

        Defaults to 7 days.
        """
        import datetime

        return self._setting(
            'CONFIRMATION_SAVE_PERIOD',
            datetime.timedelta(days=7))

    @property
    def EMAIL_VERIFICATION_URL(self):
        """
        The template to use for the email verification url.
        """
        return self._setting('EMAIL_VERIFICATION_URL', '')

    @property
    def PASSWORD_RESET_EXPIRATION(self):
        """
        The duration that a password reset token is valid for.

        Defaults to 1 hour.
        """
        import datetime

        return self._setting(
            'PASSWORD_RESET_EXPIRATION',
            datetime.timedelta(hours=1))

    @property
    def PASSWORD_RESET_URL(self):
        """
        The template to use for the password reset url.
        """
        return self._setting('PASSWORD_RESET_URL', '')

    @property
    def REGISTRATION_SERIALIZER(self):
        """
        The serializer class used for registering new users.
        """
        from django.utils.module_loading import import_string

        return import_string(
            self._setting(
                'REGISTRATION_SERIALIZER',
                'rest_email_auth.serializers.RegistrationSerializer'))


# Ugly? Guido recommends this himself ...
# http://mail.python.org/pipermail/python-ideas/2012-May/014969.html

app_settings = AppSettings()
app_settings.__name__ = __name__
sys.modules[__name__] = app_settings



try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import models


def test_generate_token():
    """
    The generate token method should use Django's provided method of
    getting a random string to create a random string.
    """
    with mock.patch(
            'rest_email_auth.models.get_random_string',
            autospec=True) as mock_random:
        models.generate_token()

    assert mock_random.call_count == 1
    assert mock_random.call_args[1] == {
        'length': 64,
    }





def test_authenticate(
        email_factory,
        user_factory,
        verified_email_auth_backend):
    """
    Passing in a verified email address and the user's password should
    successfully authenticate the user.
    """
    user = user_factory(password='password')
    email = email_factory(is_verified=True, user=user)

    assert verified_email_auth_backend.authenticate(
        email=email.email,
        password='password',
        request=None) == user


def test_authenticate_invalid_password(
        email_factory,
        user_factory,
        verified_email_auth_backend):
    """
    Providing the wrong password should cause ``None`` to be returned.
    """
    user = user_factory(password='password')
    email = email_factory(is_verified=True, user=user)

    assert verified_email_auth_backend.authenticate(
        email=email.email,
        password='notpassword',
        request=None) is None


def test_authenticate_missing_email(user_factory, verified_email_auth_backend):
    """
    If the provided email doesn't exist in the system, ``None`` should
    be returned.
    """
    user_factory(password='password')

    assert verified_email_auth_backend.authenticate(
        email='missing@example.com',
        password='password',
        request=None) is None


def test_authenticate_unverified_email(
        email_factory,
        user_factory,
        verified_email_auth_backend):
    """
    If the provided email address exists but has not been verified yet
    ``None`` should be returned.
    """
    user = user_factory(password='password')
    email = email_factory(is_verified=False, user=user)

    assert verified_email_auth_backend.authenticate(
        email=email.email,
        password='password',
        request=None) is None


def test_authenticate_username(
        email_factory,
        user_factory,
        verified_email_auth_backend):
    """
    The ``username`` parameter should act as an alias for the ``email``
    parameter.
    """
    user = user_factory(password='password')
    email = email_factory(is_verified=True, user=user)

    assert verified_email_auth_backend.authenticate(
        password='password',
        request=None,
        username=email.email) == user



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-10-25 13:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rest_email_auth', '0002_emailconfirmation'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailaddress',
            name='email',
            field=models.EmailField(max_length=255, unique=True, verbose_name='email'),
        ),
    ]



"""Views for the ``rest_email_auth`` app.
"""

from rest_framework import generics
from rest_framework.permissions import IsAuthenticated

from rest_email_auth import app_settings, serializers
from rest_email_auth.generics import SerializerSaveView


class EmailDetailView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete:
    Delete a specific email address.

    get:
    Retrieve the details of a particular email address.

    patch:
    Partially update an email address.

    Only a verified email address may be marked as the user's primary
    email.

    put:
    Update an email address.

    Only a verified emaili address may be marked as the user's primary
    email.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.EmailSerializer

    def get_queryset(self):
        """
        Get the email addresses belonging to the requesting user.

        Returns:
            A queryset containing only the email addresses owned by the
            requesting user.
        """
        return self.request.user.email_addresses.all()


class EmailListView(generics.ListCreateAPIView):
    """
    get:
    List the email addresses associated with the requesting user's
    account.

    post:
    Add a new email address to the requesting user's account.
    """
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.EmailSerializer

    def get_queryset(self):
        """
        Get the email addresses belonging to the requesting user.

        Returns:
            A queryset containing only the email addresses owned by the
            requesting user.
        """
        return self.request.user.email_addresses.all()

    def perform_create(self, serializer):
        """
        Create a new email address attached to the requesting user.

        Args:
            serializer:
                An instance of the view's serializer class containing
                the data submitted in the request.

        Returns:
            The newly created email address.
        """
        return serializer.save(user=self.request.user)


class EmailVerificationView(SerializerSaveView):
    """
    Verify a user's email address.
    """
    serializer_class = serializers.EmailVerificationSerializer


class PasswordResetRequestView(SerializerSaveView):
    """
    post:
    Request a password reset for the user with the provided email
    address. A reset token will be generated and emailed to the user.

    Notes:
    * The provided email address must be verified.
    * The operation will appear successful even if no reset email is
      sent. This is done to avoid leaking email addresses.
    """
    serializer_class = serializers.PasswordResetRequestSerializer


class PasswordResetView(SerializerSaveView):
    """
    post:
    Reset the user's password using the token that was emailed to them.
    """
    serializer_class = serializers.PasswordResetSerializer


class RegistrationView(generics.CreateAPIView):
    """
    Register a new user.
    """

    def get_serializer_class(self):
        """
        Get the serializer class used to register new users.
        """
        return app_settings.REGISTRATION_SERIALIZER


class ResendVerificationView(SerializerSaveView):
    """
    Resend an email verification to a specific address.
    """
    serializer_class = serializers.ResendVerificationSerializer



"""
A collection of reusable views.
"""

from rest_framework import generics, status
from rest_framework.response import Response


class SerializerSaveView(generics.GenericAPIView):
    """
    Generic view for saving a serializer and returning the results.
    """

    def post(self, request):
        """
        Save the provided data using the class' serializer.

        Args:
            request:
                The request being made.

        Returns:
            An ``APIResponse`` instance. If the request was successful
            the response will have a 200 status code and contain the
            serializer's data. Otherwise a 400 status code and the
            request's errors will be returned.
        """
        serializer = self.get_serializer(data=request.data)

        if serializer.is_valid():
            serializer.save()

            return Response(serializer.data)

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)





def test_get_user(base_auth_backend, user_factory):
    """
    Passing in a user's ID should return the user with that ID.
    """
    user = user_factory()

    assert base_auth_backend.get_user(user.id) == user


def test_get_user_invalid_id(base_auth_backend, db):
    """
    Passing in an ID with no corresponding user should return None.
    """
    assert base_auth_backend.get_user(1) is None



import pytest

from rest_framework import status
from rest_framework.reverse import reverse

from rest_email_auth import serializers


url = reverse('rest-email-auth:resend-verification')


@pytest.mark.django_db
@pytest.mark.integration
def test_post_resend_verification(api_client):
    """
    Sending a POST request to the endpoint should send a confirmation
    email to the provided address.
    """
    data = {'email': 'test@example.com'}

    response = api_client.post(url, data)

    assert response.status_code == status.HTTP_200_OK

    serializer = serializers.ResendVerificationSerializer(data=data)
    assert serializer.is_valid()

    assert response.data == serializer.data



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-02-05 13:53
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('rest_email_auth', '0004_emailaddress_is_primary'),
    ]

    operations = [
        migrations.CreateModel(
            name='PasswordResetToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, help_text='The time at which the password reset token was created.', verbose_name='created at')),
                ('key', models.UUIDField(default=uuid.uuid4, editable=False, help_text='The token that authorizes the user to reset their password.', verbose_name='key')),
                ('email', models.ForeignKey(editable=False, help_text='The email address used to request the password reset.', on_delete=django.db.models.deletion.CASCADE, related_name='password_reset_tokens', related_query_name='password_reset_token', to='rest_email_auth.EmailAddress', verbose_name='email address')),
            ],
            options={
                'verbose_name': 'password reset token',
                'verbose_name_plural': 'password reset tokens',
            },
        ),
    ]



try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import serializers


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_confirmation',
    autospec=True)
def test_save(mock_send_confirmation, email_factory):
    """
    Saving the serializer should send a new confirmation.
    """
    email = email_factory()

    data = {
        'email': email.email,
    }

    serializer = serializers.ResendVerificationSerializer(data=data)
    assert serializer.is_valid()

    serializer.save()

    assert mock_send_confirmation.call_count == 1


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_confirmation',
    autospec=True)
def test_save_nonexistent_email(mock_send_confirmation, db):
    """
    If the provided email address is not in the database, no action
    should be taken.
    """
    data = {
        'email': 'test@example.com',
    }

    serializer = serializers.ResendVerificationSerializer(data=data)
    assert serializer.is_valid()

    serializer.save()

    assert mock_send_confirmation.call_count == 0


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_confirmation',
    autospec=True)
def test_save_verified_email(mock_send_confirmation, email_factory):
    """
    If the provided email has already been verified, no action should be
    taken.
    """
    email = email_factory(is_verified=True)

    data = {
        'email': email.email,
    }

    serializer = serializers.ResendVerificationSerializer(data=data)
    assert serializer.is_valid()

    serializer.save()

    assert mock_send_confirmation.call_count == 0



try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import serializers


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_confirmation',
    autospec=True)
def test_create(mock_send_confirmation, user_factory):
    """
    Test creating a new email address from the serializer.

    Creating a new email address should also send a confirmation email
    for the provided address. If the user does not have a primary email
    address, the created one should be marked as the primary.
    """
    user = user_factory()
    data = {
        'email': 'test@example.com',
    }

    serializer = serializers.EmailSerializer(data=data)
    assert serializer.is_valid()

    email = serializer.save(user=user)

    assert email.email == data['email']
    assert email.is_primary

    # Make sure a confirmation email was sent
    assert mock_send_confirmation.call_count == 1


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_duplicate_notification',  # noqa
    autospec=True)
def test_create_duplicate(
        mock_duplicate_notification,
        email_factory,
        user_factory):
    """
    Attempting to add an email address that already exists should send a
    notification to the existing email.
    """
    email = email_factory()
    user = user_factory()

    data = {
        'email': email.email,
    }

    serializer = serializers.EmailSerializer(data=data)
    assert serializer.is_valid(), serializer.errors

    serializer.save(user=user)

    assert mock_duplicate_notification.call_count == 1
    assert mock_duplicate_notification.call_args[0] == (email,)


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.send_confirmation',
    autospec=True)
def test_create_non_primary(
        mock_send_confirmation,
        email_factory,
        user_factory):
    """
    If the user already has a primary email address, the created email
    should not be marked as the user's primary.
    """
    user = user_factory()
    email_factory(is_primary=True, user=user)

    data = {
        'email': 'test@example.com',
    }

    serializer = serializers.EmailSerializer(data=data)
    assert serializer.is_valid()

    email = serializer.save(user=user)

    assert email.email == data['email']
    assert not email.is_primary

    # Make sure a confirmation email was sent
    assert mock_send_confirmation.call_count == 1


def test_serialize(email_factory):
    """
    Test serializing an email address.
    """
    email = email_factory()
    serializer = serializers.EmailSerializer(email)

    expected = {
        'id': email.id,
        'created_at': email.created_at.isoformat(),
        'email': email.email,
        'is_primary': email.is_primary,
        'is_verified': email.is_verified,
    }

    assert serializer.data == expected


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.set_primary',
    autospec=True)
def test_update_is_primary(mock_set_primary, email_factory):
    """
    If an email address is verified, it should be able to be marked as
    the user's primary address.
    """
    email = email_factory(is_primary=False, is_verified=True)
    data = {
        'is_primary': True,
    }

    serializer = serializers.EmailSerializer(email, data=data, partial=True)
    assert serializer.is_valid()

    email = serializer.save()

    assert email.set_primary.call_count == 1


@mock.patch(
    'rest_email_auth.serializers.models.EmailAddress.set_primary',
    autospec=True)
def test_update_is_primary_false(mock_set_primary, email_factory):
    """
    Updating 'is_primary' to false should not call set_primary.
    """
    email = email_factory(is_primary=True, is_verified=True)
    data = {
        'is_primary': False,
    }

    serializer = serializers.EmailSerializer(email, data=data, partial=True)
    assert serializer.is_valid()

    email = serializer.save()

    assert email.set_primary.call_count == 0


def test_validate_changed_email(email_factory):
    """
    If a bound serializer attempts to change the email address of its
    instance it should not be valid.
    """
    email = email_factory(email='old@example.com')
    data = {
        'email': 'new@example.com',
    }

    serializer = serializers.EmailSerializer(email, data=data)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'email'}


def test_validate_create_primary():
    """
    Attempting to create a primary email address should not be valid. It
    should only be valid to mark a verified email address as the primary
    unless this is the user's first email.
    """
    data = {
        'email': 'test@example.com',
        'is_primary': True,
    }
    serializer = serializers.EmailSerializer(data=data)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'is_primary'}


def test_validate_make_unverified_primary(email_factory):
    """
    Attempting to mark an existing but unverified email address as the
    primary should not be valid.
    """
    email = email_factory(is_primary=False, is_verified=False)
    data = {
        'is_primary': True,
    }

    serializer = serializers.EmailSerializer(email, data=data, partial=True)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'is_primary'}



try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import serializers


def test_confirm_email(
        email_confirmation_factory,
        email_factory,
        user_factory):
    """
    If a valid confirmation key and password are provided, the email
    matching the key should be verified.
    """
    user = user_factory(password='password')
    email = email_factory(user=user)
    confirmation = email_confirmation_factory(email=email)

    data = {
        'key': confirmation.key,
        'password': 'password',
    }

    serializer = serializers.EmailVerificationSerializer(data=data)
    assert serializer.is_valid()

    serializer.save()
    email.refresh_from_db()

    expected = {
        'email': email.email,
    }

    assert serializer.data == expected
    assert email.is_verified


def test_validate_expired(email_confirmation_factory):
    """
    If the confirmation matching the provided key is expired, the
    serializer should not be valid.
    """
    confirmation = email_confirmation_factory()

    data = {
        'key': confirmation.key,
        'password': 'password',
    }

    serializer = serializers.EmailVerificationSerializer(data=data)

    with mock.patch(
            'rest_email_auth.serializers.models.EmailConfirmation.is_expired',
            new_callable=mock.PropertyMock) as mock_is_expired:
        mock_is_expired.return_value = True

        assert not serializer.is_valid()

    assert mock_is_expired.call_count == 1
    assert set(serializer.errors.keys()) == {'key'}


def test_validate_invalid_key(db):
    """
    If there is no confirmation with the provided key, the serializer
    should not be valid.
    """
    data = {
        'key': 'invalid',
        'password': 'password',
    }

    serializer = serializers.EmailVerificationSerializer(data=data)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'key'}


def test_validate_invalid_password(
        email_confirmation_factory,
        email_factory,
        user_factory):
    """
    If an invalid password is provided, the serializer should not be
    valid.
    """
    user = user_factory(password='password')
    email = email_factory(user=user)
    confirmation = email_confirmation_factory(email=email)

    data = {
        'key': confirmation.key,
        'password': 'notpassword',
    }

    serializer = serializers.EmailVerificationSerializer(data=data)

    assert not serializer.is_valid()
    assert set(serializer.errors.keys()) == {'non_field_errors'}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-31 23:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailAddress',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('email', models.EmailField(max_length=255, verbose_name='email')),
                ('is_verified', models.BooleanField(default=False, help_text='Boolean indicating if the user has verified ownership of the email address.', verbose_name='is verified')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='email_addresses', related_query_name='email_address', to=settings.AUTH_USER_MODEL, verbose_name='user')),
            ],
            options={
                'verbose_name': 'email address',
                'verbose_name_plural': 'email addresses',
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-08 15:11
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import rest_email_auth.models


class Migration(migrations.Migration):

    dependencies = [
        ('rest_email_auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailConfirmation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('key', models.CharField(default=rest_email_auth.models.generate_token, editable=False, max_length=255, verbose_name='confirmation key')),
                ('email', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='confirmations', related_query_name='confirmation', to='rest_email_auth.EmailAddress', verbose_name='email')),
            ],
            options={
                'verbose_name': 'email confirmation',
                'verbose_name_plural': 'email confirmations',
            },
        ),
    ]



from rest_framework import status
from rest_framework.reverse import reverse

from rest_email_auth import serializers


def test_delete(api_client, email_factory):
    """
    Sending a DELETE request to the view should delete the email address
    with the provided ID.
    """
    email = email_factory()
    user = email.user

    api_client.force_authenticate(user=user)

    url = reverse('rest-email-auth:email-detail', kwargs={'pk': email.pk})
    response = api_client.delete(url)

    assert response.status_code == status.HTTP_204_NO_CONTENT
    assert user.email_addresses.count() == 0


def test_get(api_client, email_factory):
    """
    Sending a GET request to the view with the ID of an email address
    owned by the requesting user should return the given email address'
    information.
    """
    email = email_factory()
    api_client.force_authenticate(user=email.user)

    url = reverse('rest-email-auth:email-detail', kwargs={'pk': email.pk})
    response = api_client.get(url)

    assert response.status_code == status.HTTP_200_OK

    serializer = serializers.EmailSerializer(email)

    assert response.data == serializer.data


def test_get_anonymous(api_client):
    """
    Sending a GET request to the view as an anonymous user should return
    a permissions error.
    """
    url = reverse('rest-email-auth:email-detail', kwargs={'pk': 1})
    response = api_client.get(url)

    assert response.status_code == status.HTTP_403_FORBIDDEN


def test_get_other_user_email(api_client, email_factory, user_factory):
    """
    Sending a GET request to the view with the ID of an email owned by
    a different user should return a 404 response.
    """
    email = email_factory()
    api_client.force_authenticate(user=user_factory())

    url = reverse('rest-email-auth:email-detail', kwargs={'pk': email.pk})
    response = api_client.get(url)

    assert response.status_code == status.HTTP_404_NOT_FOUND


def test_update(api_client, email_factory):
    """
    Sending a PATCH request to the view should allow for updating the
    email address associated with the endpoint.
    """
    email = email_factory(is_primary=False, is_verified=True)
    data = {
        'is_primary': True,
    }

    api_client.force_authenticate(user=email.user)

    url = reverse('rest-email-auth:email-detail', kwargs={'pk': email.pk})
    response = api_client.patch(url, data)

    assert response.status_code == status.HTTP_200_OK

    email.refresh_from_db()
    serializer = serializers.EmailSerializer(email)

    assert response.data == serializer.data



try:
    from unittest import mock
except ImportError:
    import mock

from django.contrib.auth import get_user_model

import pytest

from rest_email_auth import serializers


@pytest.mark.django_db
def test_create_new_user():
    """
    Saving a serializer with valid data should create a new user.
    """
    data = {
        'email': 'test@example.com',
        'password': 'password',
        'username': 'user',
    }

    serializer = serializers.RegistrationSerializer(data=data)
    assert serializer.is_valid()

    with mock.patch(
            'rest_email_auth.serializers.models.EmailAddress.send_confirmation',    # noqa
            autospec=True) as mock_send_confirmation:
        user = serializer.save()

    assert user.username == data['username']
    assert user.check_password(data['password'])
    assert user.email_addresses.count() == 1

    email = user.email_addresses.get()

    assert email.email == data['email']

    # Make sure we sent out an email confirmation
    assert mock_send_confirmation.call_count == 1


def test_register_duplicate_email(email_factory):
    """
    Attempting to register with an email that has already been verified
    should send an email to the owner of the email address notifying
    them of the registration attempt.
    """
    email = email_factory()

    data = {
        'email': email.email,
        'password': 'password',
        'username': 'user',
    }

    serializer = serializers.RegistrationSerializer(data=data)
    assert serializer.is_valid()

    with mock.patch(
            'rest_email_auth.serializers.models.EmailAddress.send_duplicate_notification',  # noqa
            autospec=True) as mock_send_duplicate_signup:
        serializer.save()

    assert get_user_model().objects.count() == 1
    assert mock_send_duplicate_signup.call_count == 1


@pytest.mark.django_db
def test_validate_password():
    """
    Validating the serializer's data should run the provided password
    through Django's default password validation system.
    """
    data = {
        'email': 'test@example.com',
        'password': 'password',
        'username': 'user',
    }

    serializer = serializers.RegistrationSerializer(data=data)

    with mock.patch(
            'rest_email_auth.serializers.password_validation.validate_password',    # noqa
            autospec=True) as mock_validate:
        assert serializer.is_valid()

    assert mock_validate.call_count == 1
    assert set(mock_validate.call_args[0]) == {data['password']}



try:
    from unittest import mock
except ImportError:
    import mock

from django.conf import settings
from django.template.loader import render_to_string

from rest_email_auth import models


def test_create(user_factory):
    """
    Test creating an email address.
    """
    email = models.EmailAddress.objects.create(
        email='test@example.com',
        is_primary=True,
        user=user_factory())

    assert not email.is_verified


def test_send_confirmation(email_factory):
    """
    Sending a confirmation should create an ``EmailConfirmation``
    instance and send a verification email.
    """
    email = email_factory()

    with mock.patch(
            'rest_email_auth.models.EmailConfirmation.send',
            autospec=True) as mock_send:
        email.send_confirmation()

    assert email.confirmations.count() == 1
    assert mock_send.call_count == 1


def test_send_duplicate_notification(email_factory, mailoutbox):
    """
    Sending a duplicate signup notification should send the user an
    email stating that their email was used to register even though
    they already have an account.
    """
    email = email_factory()
    email.send_duplicate_notification()

    context = {}
    message = render_to_string(
        context=context,
        template_name='rest_email_auth/emails/duplicate-email.txt')

    assert len(mailoutbox) == 1

    msg = mailoutbox[0]

    assert msg.body == message
    assert msg.from_email == settings.DEFAULT_FROM_EMAIL
    assert msg.subject == 'Registration Attempt'
    assert msg.to == [email.email]


def test_set_primary(email_factory):
    """
    Setting an email as the primary should update all of the user's
    other email addresses to not be the primary.
    """
    old = email_factory(is_primary=True)
    new = email_factory(is_primary=False, user=old.user)

    new.set_primary()
    old.refresh_from_db()

    assert new.is_primary
    assert not old.is_primary


def test_string_conversion(email_factory):
    """
    Converting an email address to a string should return the email
    address.
    """
    email = email_factory()

    assert str(email) == email.email



"""
Authentication backends to handle authenticating with any of a user's
email addresses.
"""

from django.contrib.auth import get_user_model

from rest_email_auth import models


class BaseBackend(object):
    """
    Base class for authentication backends.
    """

    def get_user(self, user_id):
        """
        Get a user by their ID.

        Args:
            user_id:
                The ID of the user to fetch.

        Returns:
            The user with the specified ID if they exist and ``None``
            otherwise.
        """
        try:
            return get_user_model().objects.get(id=user_id)
        except get_user_model().DoesNotExist:
            return None


class VerifiedEmailBackend(BaseBackend):
    """
    Authentication backend that only allows for the use of verified
    email addresses.
    """

    def authenticate(self, request, email=None, password=None, username=None):
        """
        Attempt to authenticate a set of credentials.

        Args:
            request:
                The request associated with the authentication attempt.
            email:
                The user's email address.
            password:
                The user's password.
            username:
                An alias for the ``email`` field. This is provided for
                compatability with Django's built in authentication
                views.

        Returns:
            The user associated with the provided credentials if they
            are valid. Returns ``None`` otherwise.
        """
        email = email or username

        try:
            email_instance = models.EmailAddress.objects.get(
                is_verified=True,
                email=email)
        except models.EmailAddress.DoesNotExist:
            return None

        user = email_instance.user

        if user.check_password(password):
            return user

        return None



try:
    from unittest import mock
except ImportError:
    import mock

from rest_framework import status
from rest_framework.reverse import reverse


url = reverse('rest-email-auth:password-reset-request')


@mock.patch(
    'rest_email_auth.views.serializers.PasswordResetRequestSerializer.save',
    autospec=True)
def test_request_reset(mock_save, api_client, email_factory):
    """
    Sending a POST request with valid data to the view should create a
    new password reset request.
    """
    email = email_factory(is_verified=True)
    data = {
        'email': email.email,
    }

    response = api_client.post(url, data)

    assert response.status_code == status.HTTP_200_OK
    assert mock_save.call_count == 1



try:
    from unittest import mock
except ImportError:
    import mock

from rest_email_auth import models, serializers


@mock.patch(
    'rest_email_auth.serializers.models.PasswordResetToken.send',
    autospec=True)
def test_save(mock_send_reset, email_factory):
    """
    Saving a serializer with valid data should send out a password reset
    email.
    """
    email = email_factory(is_verified=True)
    data = {
        'email': email.email,
    }

    serializer = serializers.PasswordResetRequestSerializer(data=data)
    assert serializer.is_valid()

    token = serializer.save()

    # Saving the serializer should not modify its data
    assert serializer.data == data

    assert token.email == email
    assert token.send.call_count == 1


def test_save_nonexistent_email(db):
    """
    If a non-existent email address is provided to the reset endpoint,
    no action should be taken.
    """
    data = {
        'email': 'fake@example.com',
    }

    serializer = serializers.PasswordResetRequestSerializer(data=data)
    assert serializer.is_valid()

    token = serializer.save()

    assert token is None
    assert models.PasswordResetToken.objects.count() == 0


def test_save_unverified_email(email_factory):
    """
    If the provided email address has not been verified yet, no action
    should be taken.
    """
    email = email_factory(is_verified=False)
    data = {
        'email': email.email,
    }

    serializer = serializers.PasswordResetRequestSerializer(data=data)
    assert serializer.is_valid()

    token = serializer.save()

    assert token is None
    assert models.PasswordResetToken.objects.count() == 0



"""Fixtures used to test the ``rest_email_auth`` app.
"""

import copy
import logging

import pytest

from rest_framework.test import APIClient, APIRequestFactory

from rest_email_auth import app_settings as application_settings, factories


logger = logging.getLogger(__name__)


class MutableAppSettings(object):
    """
    A mutable version of the application's settings.
    """

    def __init__(self, settings_obj):
        """
        Create a new mutable settings instance.

        This is done by making a copy of the REST_EMAIL_AUTH
        configuration dictionary. Since the settings fixture is rolled
        back at the end of each test, it sees that we changed the
        dictionary reference and points it back to the original.

        Args:
            settings_obj:
                The object through which Django's settings are
                accessible. We rely on the fact that this object is
                rolled back after each test.
        """
        settings_obj.REST_EMAIL_AUTH = copy.deepcopy(
            settings_obj.REST_EMAIL_AUTH)

        # Since we pass __setattr__ through to the settings dictionary
        # we have to set attributes by modifying the instance's __dict__
        # directly.
        self.__dict__['settings_obj'] = settings_obj
        self.__dict__['original'] = copy.deepcopy(settings_obj.REST_EMAIL_AUTH)

    def __getattr__(self, name):
        """
        Attempt to lookup attributes using the application's settings.

        If the attribute is not accessible through the application
        settings it falls back to the default behavior.
        """
        if hasattr(application_settings, name):
            return getattr(application_settings, name)

        return super(MutableAppSettings, self).__getattribute__(name)

    def __setattr__(self, name, value):
        """
        Setting attributes modifies the settings dictionary directly.
        """
        self.settings_obj.REST_EMAIL_AUTH[name] = value


@pytest.fixture
def api_client():
    """
    Fixture to get a client for making API requests.

    Returns:
        An instance of the APIClient provided by DRF.
    """
    return APIClient()


@pytest.fixture
def api_rf():
    """
    Fixture to get a factory for creating API requests.

    Returns:
        An instance of the request factory provided by DRF.
    """
    return APIRequestFactory()


@pytest.fixture
def app_settings(settings):
    """
    Fixture to get a mutable copy of the application's settings.

    Any changes are rolled back at the end of the test.
    """
    return MutableAppSettings(settings)


@pytest.fixture
def email_confirmation_factory(db):
    """
    Fixture to get the factory used to create email confirmations.

    Returns:
        The factory used to create ``EmailConfirmation`` instances.
    """
    return factories.EmailConfirmationFactory


@pytest.fixture
def email_factory(db):
    """
    Fixture to get the factory used to create email addresses.

    Returns:
        The factory used to create ``EmailAddress`` instances.
    """
    return factories.EmailFactory


@pytest.fixture
def password_reset_token_factory(db):
    """
    Fixture to get the factory usedd to create password reset tokens.

    Returns:
        The factory used to create ``PasswordResetToken`` instances.
    """
    return factories.PasswordResetTokenFactory


@pytest.fixture
def user_factory(db):
    """
    Fixture to get the factory used to create users.

    Returns:
        The factory used to create Django user instances.
    """
    return factories.UserFactory



try:
    from unittest import mock
except ImportError:
    import mock

import pytest

from rest_framework import status

from rest_email_auth import app_settings, views


registration_view = views.RegistrationView.as_view()


@mock.patch(
    'rest_email_auth.views.app_settings.__class__.REGISTRATION_SERIALIZER',
    new_callable=mock.PropertyMock)
def test_get_serializer_class(mock_registration_serializer):
    """
    The view should use the serializer defined in the app's settings.
    """
    view = views.RegistrationView()
    expected = mock_registration_serializer.return_value

    assert view.get_serializer_class() == expected


@pytest.mark.django_db
def test_register(api_rf):
    """
    Sending a POST request with valid data to the view should register a
    new user.
    """
    data = {
        'email': 'test@example.com',
        'password': 'password',
        'username': 'user',
    }

    serializer = app_settings.REGISTRATION_SERIALIZER(data=data)
    assert serializer.is_valid()

    request = api_rf.post('/', data)
    response = registration_view(request)

    assert response.status_code == status.HTTP_201_CREATED
    assert response.data == serializer.data



from setuptools import setup, find_packages


def get_description():
    with open('README.rst') as f:
        return f.read()


setup(
    # Package meta-data
    name='django-rest-email-auth',
    version='0.4.2',
    description='Django app for email based authentication and registration.',
    long_description=get_description(),
    author='Chathan Driehuys',
    author_email='cdriehuys@gmail.com',
    url='https://github.com/cdriehuys/django-rest-email-auth',
    license='MIT',

    # Additional classifiers for PyPI
    classifiers=[
        'Development Status :: 3 - Alpha',

        # Supported versions of Django
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',

        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',

        # Supported versions of Python
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # Include the actual source code
    packages=find_packages(),

    # Dependencies
    install_requires=[
        'Django >= 1.10',
        'djangorestframework >= 3.0',
    ])



from django.db import models
from django.utils import timezone

from rest_email_auth import app_settings


class ValidPasswordResetTokenManager(models.Manager):
    """
    Manager for getting only valid password reset tokens.

    Valid tokens are those that have not yet expired.
    """

    def get_queryset(self):
        """
        Return all unexpired password reset tokens.
        """
        oldest = timezone.now() - app_settings.PASSWORD_RESET_EXPIRATION
        queryset = super(ValidPasswordResetTokenManager, self).get_queryset()

        return queryset.filter(created_at__gt=oldest)



"""Serializers for the ``rest_email_auth`` app.

The serializers handle the conversion of data between the JSON or form
data the API receives and native Python datatypes.
"""

import logging

from django.contrib.auth import get_user_model, password_validation
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers

from rest_email_auth import models


logger = logging.getLogger(__name__)


class EmailSerializer(serializers.ModelSerializer):
    """
    Serializer for email addresses.
    """

    class Meta(object):
        extra_kwargs = {
            'email': {
                # We remove the autogenerated 'unique' validator to
                # avoid leaking email addresses.
                'validators': [],
            },
        }
        fields = ('id', 'created_at', 'email', 'is_primary', 'is_verified')
        model = models.EmailAddress
        read_only_fields = ('is_verified',)

    def create(self, validated_data):
        """
        Create a new email and send a confirmation to it.

        Returns:
            The newly creating ``EmailAddress`` instance.
        """
        email_query = models.EmailAddress.objects.filter(
            email=self.validated_data['email'])

        if email_query.exists():
            email = email_query.get()

            email.send_duplicate_notification()
        else:
            email = super(EmailSerializer, self).create(validated_data)
            email.send_confirmation()

            user = validated_data.get('user')
            query = models.EmailAddress.objects.filter(
                is_primary=True,
                user=user)

            if not query.exists():
                email.set_primary()

        return email

    def update(self, instance, validated_data):
        """
        Update the instance the serializer is bound to.

        Args:
            instance:
                The instance the serializer is bound to.
            validated_data:
                The data to update the serializer with.

        Returns:
            The updated instance.
        """
        is_primary = validated_data.pop('is_primary', False)

        instance = super(EmailSerializer, self).update(
            instance,
            validated_data)

        if is_primary:
            instance.set_primary()

        return instance

    def validate_email(self, email):
        """
        Validate the provided email address.

        Returns:
            The validated email address.

        Raises:
            serializers.ValidationError:
                If the serializer is bound and the provided email
                doesn't match the existing address.
        """
        if self.instance and email and self.instance.email != email:
            raise serializers.ValidationError(
                _("Existing emails may not be edited. Create a new one "
                  "instead."))

        return email

    def validate_is_primary(self, is_primary):
        """
        Validate the provided 'is_primary' parameter.

        Returns:
            The validated 'is_primary' value.

        Raises:
            serializers.ValidationError:
                If the user attempted to mark an unverified email as
                their primary email address.
        """
        # TODO: Setting 'is_primary' to 'False' should probably not be
        #       allowed.
        if is_primary and not (self.instance and self.instance.is_verified):
            raise serializers.ValidationError(
                _("Unverified email addresses may not be used as the primary "
                  "address."))

        return is_primary


class EmailVerificationSerializer(serializers.Serializer):
    """
    Serializer for verifying an email address.
    """
    email = serializers.EmailField(read_only=True)
    key = serializers.CharField(write_only=True)
    password = serializers.CharField(
        style={'input_type': 'password'},
        write_only=True)

    def save(self):
        """
        Confirm the email address matching the confirmation key.
        """
        self._confirmation.confirm()

    def validate(self, data):
        """
        Validate the provided data.

        Returns:
            dict:
                The validated data.

        Raises:
            serializers.ValidationError:
                If the provided password is invalid.
        """
        user = self._confirmation.email.user

        if not user.check_password(data['password']):
            raise serializers.ValidationError(
                _('The provided password is invalid.'))

        # Add email to returned data
        data['email'] = self._confirmation.email.email

        return data

    def validate_key(self, key):
        """
        Validate the provided confirmation key.

        Returns:
            str:
                The validated confirmation key.

        Raises:
            serializers.ValidationError:
                If there is no email confirmation with the given key or
                the confirmation has expired.
        """
        try:
            confirmation = models.EmailConfirmation.objects.select_related(
                'email__user').get(key=key)
        except models.EmailConfirmation.DoesNotExist:
            raise serializers.ValidationError(
                _('The provided verification key is invalid.'))

        if confirmation.is_expired:
            raise serializers.ValidationError(
                _('That verification code has expired.'))

        # Cache confirmation instance
        self._confirmation = confirmation

        return key


class PasswordResetRequestSerializer(serializers.Serializer):
    """
    Serializer for requesting a password reset.
    """
    email = serializers.EmailField(
        help_text=_("The email address to send the password reset to."))

    def save(self):
        """
        Send out a password reset if the provided data is valid.

        If the provided email address exists and is verified, a reset
        email is sent to the address.

        Returns:
            The password reset token if it was returned and ``None``
            otherwise.
        """
        try:
            email = models.EmailAddress.objects.get(
                email=self.validated_data['email'],
                is_verified=True)
        except models.EmailAddress.DoesNotExist:
            return None

        token = models.PasswordResetToken.objects.create(email=email)
        token.send()

        return token


class PasswordResetSerializer(serializers.Serializer):
    """
    Serializer for reseting a user's password.
    """
    key = serializers.UUIDField(
        help_text=_('The key received by the user in the password reset '
                    'email.'),
        write_only=True)
    password = serializers.CharField(
        help_text=_("The user's new password."),
        style={'input_type': 'password'},
        write_only=True)

    def save(self):
        """
        Reset the user's password if the provided information is valid.
        """
        token = models.PasswordResetToken.objects.get(
            key=self.validated_data['key'])

        token.email.user.set_password(self.validated_data['password'])
        token.email.user.save()

        logger.info("Reset password for %s", token.email.user)

        token.delete()

    def validate_key(self, key):
        """
        Validate the provided reset key.

        Returns:
            The validated key.

        Raises:
            serializers.ValidationError:
                If the provided key does not exist.
        """
        if not models.PasswordResetToken.valid_tokens.filter(key=key).exists():
            raise serializers.ValidationError(
                _("The provided reset token does not exist, or is expired."))

        return key

    def validate_password(self, password):
        """
        Validate the provided password by running it through Django's
        password validation system.

        Returns:
            The validated password.

        Raises:
            ValidationError:
                If the provided password does not pass the configured
                password validators.
        """
        password_validation.validate_password(password)

        return password


class RegistrationSerializer(serializers.ModelSerializer):
    """
    Serializer for registering new users.
    """
    email = serializers.EmailField()

    class Meta(object):
        extra_kwargs = {
            'password': {
                'style': {'input_type': 'password'},
                'write_only': True,
            },
        }
        fields = (get_user_model().USERNAME_FIELD, 'email', 'password')
        model = get_user_model()

    def create(self, validated_data):
        """
        Create a new user from the data passed to the serializer.

        If the provided email has not been verified yet, the user is
        created and a verification email is sent to the address.
        Otherwise we send a notification to the email address that
        someone attempted to register with an email that's already been
        verified.

        Args:
            validated_data (dict):
                The data passed to the serializer after it has been
                validated.

        Returns:
            A new user created from the provided data.
        """
        email = validated_data.pop('email')
        password = validated_data.pop('password')

        # We don't save the user instance yet in case the provided email
        # address already exists.
        user = get_user_model()(**validated_data)
        user.set_password(password)

        # We set an ephemeral email property so that it is included in
        # the data returned by the serializer.
        user.email = email

        email_query = models.EmailAddress.objects.filter(email=email)

        if email_query.exists():
            existing_email = email_query.get()
            existing_email.send_duplicate_notification()
        else:
            user.save()

            email_instance = models.EmailAddress.objects.create(
                email=email,
                user=user)
            email_instance.send_confirmation()

        return user

    def validate_password(self, password):
        """
        Validate the provided password.

        Args:
            password (str):
                The password provided by the user.

        Returns:
            str:
                The validated password.

        Raises:
            ValidationError:
                If the provided password doesn't pass Django's provided
                password validation.
        """
        password_validation.validate_password(password)

        return password


class ResendVerificationSerializer(serializers.Serializer):
    """
    Serializer for resending a verification email.
    """
    email = serializers.EmailField()

    def save(self):
        """
        Resend a verification email to the provided address.

        If the provided email is already verified no action is taken.
        """
        try:
            email = models.EmailAddress.objects.get(
                email=self.validated_data['email'],
                is_verified=False)

            logger.debug(
                'Resending verification email to %s',
                self.validated_data['email'])

            email.send_confirmation()
        except models.EmailAddress.DoesNotExist:
            logger.debug(
                'Not resending verification email to %s because the address '
                "doesn't exist in the database.",
                self.validated_data['email'])



import datetime

try:
    from unittest import mock
except ImportError:
    import mock

from django.template.loader import render_to_string

from rest_email_auth import models


def test_confirm(email_confirmation_factory, email_factory):
    """
    Confirming a confirmation should mark the associated email address
    as verified.
    """
    email = email_factory()
    confirmation = email_confirmation_factory(email=email)

    confirmation.confirm()
    email.refresh_from_db()

    assert email.is_verified


def test_create_email_confirmation(email_factory):
    """
    Test creating a new email confirmation.
    """
    models.EmailConfirmation.objects.create(
        email=email_factory())


def test_is_expired(email_confirmation_factory):
    """
    If it has been more than a day since the confirmation was created it
    should be marked as expired.
    """
    confirmation = email_confirmation_factory()
    future_time = confirmation.created_at + datetime.timedelta(days=2)

    with mock.patch(
            'rest_email_auth.models.timezone.now',
            autospec=True,
            return_value=future_time):
        assert confirmation.is_expired


def test_is_expired_unexpired(email_confirmation_factory):
    """
    If it has been less than a day since the confirmation was created it
    should not be marked as expired.
    """
    confirmation = email_confirmation_factory()

    assert not confirmation.is_expired


def test_send(email_confirmation_factory, mailoutbox, settings):
    """
    Sending the confirmation should send an email to the associated
    email address.
    """
    settings.REST_EMAIL_AUTH = {
        'EMAIL_VERIFICATION_URL': 'https://example.com/verify?key={key}',
    }

    confirmation = email_confirmation_factory()
    confirmation.send()

    url = settings.REST_EMAIL_AUTH['EMAIL_VERIFICATION_URL'].format(
        key=confirmation.key)

    context = {
        'verification_url': url,
    }
    expected_content = render_to_string(
        context=context,
        template_name='rest_email_auth/emails/verify-email.txt')

    assert len(mailoutbox) == 1

    msg = mailoutbox[0]

    assert msg.body == expected_content
    assert msg.from_email == settings.DEFAULT_FROM_EMAIL
    assert msg.subject == 'Please Verify Your Email Address'
    assert msg.to == [confirmation.email.email]



import datetime

from django.template.loader import render_to_string

from rest_email_auth import models


def test_create(email_factory):
    """
    Test creating a new password reset token.
    """
    models.PasswordResetToken.objects.create(email=email_factory())


def test_string_conversion(password_reset_token_factory):
    """
    Converting a password reset token to a string should return
    information about the token's owner.
    """
    token = password_reset_token_factory()
    expected = "Password reset token for '{}'".format(token.email.user)

    assert str(token) == expected


# For some reason, if 'mailoutbox' is listed before
# 'password_reset_token_factory', the test breaks.
def test_send(password_reset_token_factory, mailoutbox, settings):
    """
    Sending the token should send an email to the address associated
    with the token.
    """
    settings.REST_EMAIL_AUTH = {
        'PASSWORD_RESET_URL': 'https://example.com/reset/{key}',
    }

    token = password_reset_token_factory()
    token.send()

    url = settings.REST_EMAIL_AUTH['PASSWORD_RESET_URL'].format(
        key=token.key)

    context = {
        'reset_url': url,
    }
    expected_content = render_to_string(
        context=context,
        template_name='rest_email_auth/emails/reset-password.txt')

    assert len(mailoutbox) == 1

    msg = mailoutbox[0]

    assert msg.body == expected_content
    assert msg.from_email == settings.DEFAULT_FROM_EMAIL
    assert msg.subject == 'Reset Your Password'
    assert msg.to == [token.email.email]


def test_valid_tokens(password_reset_token_factory, app_settings):
    """
    The valid tokens queryset should include unexpired tokens.
    """
    app_settings.PASSWORD_RESET_EXPIRATION = datetime.timedelta(hours=1)

    token = password_reset_token_factory()

    assert list(models.PasswordResetToken.valid_tokens.all()) == [token]


def test_valid_tokens_expired(password_reset_token_factory, app_settings):
    """
    The valid tokens queryset should not include expired tokens.
    """
    app_settings.PASSWORD_RESET_EXPIRATION = datetime.timedelta()

    password_reset_token_factory()

    assert not models.PasswordResetToken.valid_tokens.exists()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-01-31 20:25
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rest_email_auth', '0003_unique_emails'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailaddress',
            name='is_primary',
            field=models.BooleanField(default=False, help_text="Boolean indicating if the email is the user's primary address.", verbose_name='is primary'),
        ),
    ]



"""
Fixtures for testing authentication backends.
"""

import pytest

from rest_email_auth import authentication


@pytest.fixture(scope='session')
def base_auth_backend():
    """
    Return an instance of the base authentication backend.
    """
    return authentication.BaseBackend()


@pytest.fixture(scope='session')
def verified_email_auth_backend():
    """
    Return an instance of the authentication backend that only accepts
    verified emails.
    """
    return authentication.VerifiedEmailBackend()



"""Admin configurations for the ``rest_email_auth`` app.
"""

from django.contrib import admin
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _

from rest_email_auth import models


@admin.register(models.EmailAddress)
class EmailAddressAdmin(admin.ModelAdmin):
    """
    Admin for ``EmailAddress`` instances.
    """
    fields = ('user', 'email', 'is_verified', 'created_at')
    list_display = ('email', 'user', 'is_verified', 'created_at')
    list_filter = ('is_verified',)
    readonly_fields = ('created_at',)
    search_fields = ('email',)


@admin.register(models.EmailConfirmation)
class EmailConfirmationAdmin(admin.ModelAdmin):
    """
    Admin for ``EmailConfirmation`` instances.
    """
    fields = ('email', 'key', 'created_at')
    list_display = ('email', 'created_at')
    readonly_fields = ('email', 'key', 'created_at')
    search_fields = ('email__email',)


@admin.register(models.PasswordResetToken)
class PasswordResetTokenAdmin(admin.ModelAdmin):
    """
    Admin for ``PasswordResetToken`` instances.
    """
    fields = ('email', 'key', 'created_at')
    list_display = ('get_user', 'email', 'created_at')
    readonly_fields = fields
    search_fields = (get_user_model().USERNAME_FIELD, 'email__email')

    def get_user(self, obj):
        """
        Get the user that owns the password reset token.
        """
        return obj.email.user
    get_user.admin_order_field = 'email__user'
    get_user.short_description = _('user')



import datetime
import logging
import uuid

from django.conf import settings
from django.core import mail
from django.db import models, transaction
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.translation import ugettext_lazy as _

from rest_email_auth import app_settings, managers


logger = logging.getLogger(__name__)


def generate_token():
    """
    Get a random 64 character string.

    Returns:
        str:
            A random 64 character string.
    """
    return get_random_string(length=64)


class EmailAddress(models.Model):
    """
    A user's email address.
    """
    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('created at'))
    email = models.EmailField(
        max_length=255,
        unique=True,
        verbose_name=_('email'))
    is_primary = models.BooleanField(
        default=False,
        help_text=_("Boolean indicating if the email is the user's primary "
                    "address."),
        verbose_name=_('is primary'))
    is_verified = models.BooleanField(
        default=False,
        help_text=_('Boolean indicating if the user has verified ownership of '
                    'the email address.'),
        verbose_name=_('is verified'))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='email_addresses',
        related_query_name='email_address',
        verbose_name=_('user'))

    class Meta(object):
        verbose_name = _('email address')
        verbose_name_plural = _('email addresses')

    def __str__(self):
        """
        Get a string representation of the email address.

        Returns:
            str:
                A text version of the email address. Ex:
                ``test@example.com``.
        """
        return self.email

    def send_confirmation(self):
        """
        Send a verification email for the email address.
        """
        confirmation = EmailConfirmation.objects.create(email=self)
        confirmation.send()

    def send_duplicate_notification(self):
        """
        Send a notification about a duplicate signup.
        """
        context = {}
        message = render_to_string(
            context=context,
            template_name='rest_email_auth/emails/duplicate-email.txt')

        mail.send_mail(
            from_email=None,
            message=message,
            recipient_list=[self.email],
            subject=_('Registration Attempt'))

        logger.info('Sent duplicate email notification to: %s', self.email)

    def set_primary(self):
        """
        Set this email address as the user's primary email.
        """
        query = EmailAddress.objects.filter(is_primary=True, user=self.user)
        query = query.exclude(pk=self.pk)

        # The transaction is atomic so there is never a gap where a user
        # has no primary email address.
        with transaction.atomic():
            query.update(is_primary=False)

            self.is_primary = True
            self.save()

        logger.info(
            'Set %s as the primary email address for %s.',
            self.email,
            self.user)


class EmailConfirmation(models.Model):
    """
    Model to store a token used to verify an email address.
    """
    created_at = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('created at'))
    email = models.ForeignKey(
        'rest_email_auth.EmailAddress',
        on_delete=models.CASCADE,
        related_name='confirmations',
        related_query_name='confirmation',
        verbose_name=_('email'))
    key = models.CharField(
        default=generate_token,
        editable=False,
        max_length=255,
        verbose_name=_('confirmation key'))

    class Meta(object):
        verbose_name = _('email confirmation')
        verbose_name_plural = _('email confirmations')

    def confirm(self):
        """
        Mark the instance's email as verified.
        """
        self.email.is_verified = True
        self.email.save()

        logger.info('Verified email address: %s', self.email.email)

    @property
    def is_expired(self):
        """
        Determine if the confirmation has expired.

        Returns:
            bool:
                ``True`` if the confirmation has expired and ``False``
                otherwise.
        """
        expiration_time = self.created_at + datetime.timedelta(days=1)

        return timezone.now() > expiration_time

    def send(self):
        """
        Send a verification email to the user.
        """
        context = {
            'verification_url': app_settings.EMAIL_VERIFICATION_URL.format(
                key=self.key),
        }
        message = render_to_string(
            context=context,
            template_name='rest_email_auth/emails/verify-email.txt')

        mail.send_mail(
            from_email=None,
            message=message,
            recipient_list=[self.email.email],
            subject=_('Please Verify Your Email Address'))


class PasswordResetToken(models.Model):
    """
    Store a token that can be used to reset a user's password.
    """
    created_at = models.DateTimeField(
        auto_now_add=True,
        help_text=_('The time at which the password reset token was created.'),
        verbose_name=_('created at'))
    email = models.ForeignKey(
        'rest_email_auth.EmailAddress',
        editable=False,
        help_text=_("The email address used to request the password reset."),
        on_delete=models.CASCADE,
        related_name='password_reset_tokens',
        related_query_name='password_reset_token',
        verbose_name=_('email address'))
    key = models.UUIDField(
        default=uuid.uuid4,
        editable=False,
        help_text=_("The token that authorizes the user to reset their "
                    "password."),
        verbose_name=_('key'))

    # Custom managers. We have to explicitly define the default manager
    # so that it doesn't get overwritten.
    objects = models.Manager()
    valid_tokens = managers.ValidPasswordResetTokenManager()

    class Meta:
        verbose_name = _('password reset token')
        verbose_name_plural = _('password reset tokens')

    def __str__(self):
        """
        Get a string representation of the instance.

        Returns:
            Information about the token's owner.
        """
        return "Password reset token for '{}'".format(self.email.user)

    def send(self):
        """
        Send the password reset token to the user.
        """
        context = {
            'reset_url': app_settings.PASSWORD_RESET_URL.format(
                key=self.key),
        }
        message = render_to_string(
            context=context,
            template_name='rest_email_auth/emails/reset-password.txt')

        mail.send_mail(
            from_email=None,
            message=message,
            recipient_list=[self.email.email],
            subject=_('Reset Your Password'))

        logger.info('Sent password reset email to %s', self.email)
